#!/usr/bin/env scheme-script
;; -*- mode: scheme; coding: utf-8 -*- !#
;; FAT file system driver in Scheme
;; Copyright © 2019, 2020 Göran Weinholt <goran@weinholt.se>

;; fs-fatfs is free software: you can redistribute it and/or modify
;; it under the terms of the GNU Lesser General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.

;; fs-fatfs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU Lesser General Public License for more details.

;; You should have received a copy of the GNU Lesser General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.
#!r6rs

(import
  (rnrs (6))
  (srfi :64)
  (fs fatfs)
  (fs fatfs port-device))

(define (ls/R x path)
  (let ((rev-path (reverse path)))
    (let ((dir (fatfs-open-directory x path)))
      (let lp ()
        (let ((stat (fatfs-read-directory dir)))
          (if (eof-object? stat)
              (fatfs-close-directory dir)
              (let ((name (fatfs-file-info:filename stat))
                    (lfn (fatfs-file-info:long-filename stat)))
                (let ((full-name (reverse (cons name rev-path))))
                  (write (list full-name lfn (fatfs-file-info:size stat)))
                  (newline)
                  (when (and (fatfs-file-info-directory? stat)
                             (not (member (fatfs-file-info:filename stat) '("." ".."))))
                    (ls/R x full-name)))
                (lp))))))))

(define (directory-files fs dirname)
  (let ((dir (fatfs-open-directory fs dirname)))
    (let lp ((ret '()))
      (let ((info (fatfs-read-directory dir)))
        (cond ((eof-object? info)
               (fatfs-close-directory dir)
               (reverse ret))
              (else
               (lp (cons (or (fatfs-file-info:long-filename info)
                             (fatfs-file-info:filename info))
                         ret))))))))

(define (touch fs path)
  (call-with-port (fatfs-open-file fs path (fxior fatfs-open/write fatfs-open/create))
    values))

(let ((x (open-fatfs (open-file-fatfs-device "test.img" 512))))
  (ls/R x '())

  (assert (fatfs-file-exists? x '()))
  (assert (fatfs-file-exists? x '("README.md")))
  (assert (fatfs-file-exists? x '("readme.md")))
  (assert (not (fatfs-file-exists? x '(""))))
  (assert (not (fatfs-file-exists? x '("README.md" "x"))))

  (assert (fatfs-file-exists? x '("run-te~1.sh")))
  (assert (fatfs-file-exists? x '("run-tests.sh")))
  (fatfs-delete-file x '("run-tests.sh"))
  (assert (not (fatfs-file-exists? x '("run-te~1.sh"))))
  (assert (not (fatfs-file-exists? x '("run-tests.sh"))))

  (call-with-port (fatfs-open-file x '("README.md") fatfs-open/read)
    (lambda (port)
      (write (utf8->string (get-bytevector-all port)))
      (newline)
      (flush-output-port (current-output-port))))

  (call-with-port (transcoded-port (fatfs-open-file x '("README.md") fatfs-open/read)
                                   (native-transcoder))
    (lambda (port)
      (write port)
      (newline)
      (let lp ()
        (let ((line (get-line port)))
          (write line)
          (newline)
          (unless (eof-object? line)
            (lp))))))
  (flush-output-port (current-output-port))

  #;
  (call-with-port (transcoded-port (fatfs-open-file x '("fatfs.sls"))
                                   (native-transcoder))
    (lambda (port)
      (write port) (newline)
      (let ((data (get-string-all port)))
        (display "fatfs.sls: ")
        (write data)
        (newline))))
  (newline)

  (display "seek test\n")
  (flush-output-port (current-output-port))
  (call-with-port (fatfs-open-file x '("README.md") fatfs-open/read)
    (lambda (port)
      (let ((b (get-u8 port)))
        (assert (= (port-position port) 1))
        (set-port-position! port 0)
        (assert (eqv? (get-u8 port) b)))))

  (display "truncation test\n")
  (flush-output-port (current-output-port))
  (let* ((fn '("README.md"))
         (len (fatfs-file-info:size (fatfs-file-info x fn))))
    (fatfs-truncate-file x fn len)
    (assert (eqv? len (fatfs-file-info:size (fatfs-file-info x fn))))
    (fatfs-truncate-file x fn 4096)
    (assert (eqv? 4096 (fatfs-file-info:size (fatfs-file-info x fn))))
    (fatfs-truncate-file x fn 700000)
    (assert (eqv? 700000 (fatfs-file-info:size (fatfs-file-info x fn))))
    (fatfs-truncate-file x fn 1024)
    (assert (eqv? 1024 (fatfs-file-info:size (fatfs-file-info x fn))))
    (fatfs-truncate-file x fn 0)
    (assert (eqv? 0 (fatfs-file-info:size (fatfs-file-info x fn))))
    (fatfs-truncate-file x fn len)
    (assert (eqv? len (fatfs-file-info:size (fatfs-file-info x fn))))
    (fatfs-set-file-mode x fn (fxior FATFS-ARCHIVE))
    (let ((adate 0)
          (wdate (fatfs-raw-date 2020 7 5))
          (wtime (fatfs-raw-time 21 53 12))
          (cdate 0)
          (ctime 0)
          (ctime-ms 0))
      (fatfs-set-file-timespecs x fn adate wdate wtime cdate ctime ctime-ms))
    (fatfs-truncate-file x fn 0)
    (assert (eqv? 0 (fatfs-file-info:size (fatfs-file-info x fn))))
    (display "truncation ok\n")
    (flush-output-port (current-output-port))

    (display "read/write test\n")
    (flush-output-port (current-output-port))
    (call-with-port (fatfs-open-file x '("README.md") fatfs-open/read+write)
      (lambda (port)
        (write port)
        (newline)
        (put-bytevector port (make-bytevector 512 (char->integer #\A)))
        (put-bytevector port (make-bytevector 512 (char->integer #\B)))
        (set-port-position! port 1)
        (put-bytevector port (string->utf8 "Hello, world!\n"))
        (set-port-position! port 513)
        (put-bytevector port (string->utf8 "1234"))
        (set-port-position! port 2048)
        (put-u8 port (char->integer #\U))))
    (display "read/write test ok\n")
    (flush-output-port (current-output-port)))

  (call-with-port (transcoded-port
                   (fatfs-open-file x '("NEW.FIL")
                                    (fxior fatfs-open/write
                                           fatfs-open/create))
                   (make-transcoder (utf-8-codec) (eol-style crlf)))
    (lambda (port)
      (write port) (newline)
      (put-string port "Hello, world!\n")))

  (touch x '("1lc.lc"))
  (touch x '("2lc.UC"))
  (touch x '("3UC.lc"))
  (touch x '("4UC.UC"))
  (touch x '("1.lc"))
  (touch x '("1.UC"))
  (touch x '("lc.1"))
  (touch x '("UC.1"))

  (close-fatfs x))

(flush-output-port (current-output-port))

(let ((fs (open-fatfs (open-file-fatfs-device "test.img" 512))))
  (call-with-port (transcoded-port (fatfs-open-file fs '("HELLO.TXT")
                                                    (fxior fatfs-open/write fatfs-open/create))
                                   (native-transcoder))
    (lambda (port)
      (put-string port "Hello, world!\n")))

  (test-begin "delete-directory")
  (test-assert (fatfs-file-exists? fs '("empty")))
  (fatfs-delete-directory fs '("Empty"))
  (test-assert (not (fatfs-file-exists? fs '("empty"))))
  (test-error (fatfs-delete-directory fs '("Empty")))
  (test-assert (not (fatfs-file-exists? fs '("empty"))))
  (test-end)

  (test-begin "delete-nonempty-directory")
  (test-assert (fatfs-file-exists? fs '("unempty")))
  (test-error (fatfs-delete-directory fs '("unempty")))
  (test-assert (fatfs-file-exists? fs '("unempty" "readme.md")))
  (fatfs-delete-file fs '("unempty" "readme.md"))
  (test-assert (not (fatfs-file-exists? fs '("unempty" "readme.md"))))
  (fatfs-delete-directory fs '("unempty"))
  (test-assert (not (fatfs-file-exists? fs '("unempty"))))
  (test-end)

  (test-begin "create-directory")
  (fatfs-create-directory fs '("newdir"))
  (test-assert (fatfs-file-exists? fs '("newdir")))
  (call-with-port (transcoded-port (fatfs-open-file fs '("newdir" "newfile.txt")
                                                    (fxior fatfs-open/write fatfs-open/create))
                                   (make-transcoder (latin-1-codec) (eol-style crlf)))
    (lambda (port)
      (put-string port "The lemonade will be ¢10.\n")))
  (test-end)

  (test-begin "create-subdir")
  (fatfs-create-directory fs '("dir"))
  (fatfs-create-directory fs '("dir" "subdir"))
  (test-assert (fatfs-file-exists? fs '("dir")))
  (test-assert (fatfs-file-exists? fs '("dir" "subdir")))
  (call-with-port (transcoded-port (fatfs-open-file fs '("dir" "subdir" "readme.txt")
                                                    (fxior fatfs-open/write fatfs-open/create))
                                   (make-transcoder (latin-1-codec) (eol-style crlf)))
    (lambda (port)
      (put-string port "File in a subdirectory!")))
  (test-assert (fatfs-file-exists? fs '("dir" "subdir" "README.TXT")))
  (test-end)

  (let ((files (directory-files fs '())))
    (test-begin "create-lfn")
    (test-assert (not (fatfs-file-exists? fs '("Long filename.txt"))))
    (touch fs '("Long filename.txt"))
    (test-assert (fatfs-file-exists? fs '("Long filename.txt")))

    (let ((files^ (directory-files fs '())))
      (test-equal (append files '("Long filename.txt")) files^))

    (touch fs '("Long filename 2.txt"))
    (test-assert (fatfs-file-exists? fs '("Long filename 2.txt")))
    (let ((files^ (directory-files fs '())))
      (test-equal (append files '("Long filename.txt" "Long filename 2.txt")) files^))

    (touch fs '("Long filename 3.txt"))
    (touch fs '("Long filename 4.txt"))
    (touch fs '("Long filename 5.txt"))
    (touch fs '("Long filename 6.txt"))
    (touch fs '("Long filename 7.txt"))

    (fatfs-create-directory fs '("Program Files"))
    (do ((i 0 (fx+ i 1)))
        ((fx=? i 250))
      (let ((fn (list "Program Files"
                      (string-append (make-string i #\a) ".["))))
        (touch fs fn)
        (fatfs-delete-file fs fn)))
    (test-end)

    (test-begin "rename")
    (call-with-port (transcoded-port (fatfs-open-file fs '("OLDNAME.TXT")
                                                      (fxior fatfs-open/write fatfs-open/create))
                                     (make-transcoder (latin-1-codec) (eol-style crlf)))
      (lambda (port)
        (put-string port "This file will be called NEWNAME.TXT.\n")))
    (test-assert (fatfs-file-exists? fs '("OLDNAME.TXT")))
    (test-assert (not (fatfs-file-exists? fs '("NEWNAME.TXT"))))
    (fatfs-rename-file fs '("OLDNAME.TXT") '("NEWNAME.TXT"))
    (test-assert (not (fatfs-file-exists? fs '("OLDNAME.TXT"))))
    (test-assert (fatfs-file-exists? fs '("NEWNAME.TXT")))
    (test-end)

    (test-begin "open")
    (call-with-port (fatfs-open-file fs '("OPEN.TXT") (fxior fatfs-open/write fatfs-open/create))
      (lambda (port) (put-bytevector port #vu8(0 1 2 3 4))))
    (test-error (fatfs-open-file fs '("OPEN.TXT") (fxior fatfs-open/write fatfs-open/exclusive)))
    (call-with-port (fatfs-open-file fs '("OPEN.TXT") (fxior fatfs-open/write))
      (lambda (port) (put-bytevector port #vu8(255))))
    (test-equal #vu8(255 1 2 3 4)
                (call-with-port (fatfs-open-file fs '("OPEN.TXT") (fxior fatfs-open/read))
                  get-bytevector-all))
    (call-with-port (fatfs-open-file fs '("OPEN.TXT") (fxior fatfs-open/write fatfs-open/append))
      (lambda (port) (put-bytevector port #vu8(123))))
    (test-equal #vu8(255 1 2 3 4 123)
                (call-with-port (fatfs-open-file fs '("OPEN.TXT") (fxior fatfs-open/read))
                  get-bytevector-all))
    (call-with-port (fatfs-open-file fs '("OPEN.TXT") (fxior fatfs-open/write fatfs-open/truncate))
      (lambda (port) (put-bytevector port #vu8(5))))
    (test-equal #vu8(5)
                (call-with-port (fatfs-open-file fs '("OPEN.TXT") fatfs-open/read)
                  get-bytevector-all))
    (call-with-port (fatfs-open-file fs '("OPEN.TXT") (fxior fatfs-open/write fatfs-open/truncate fatfs-open/append))
      (lambda (port) (put-bytevector port #vu8(5))))
    (test-equal #vu8(5)
                (call-with-port (fatfs-open-file fs '("OPEN.TXT") fatfs-open/read)
                  get-bytevector-all))
    (fatfs-delete-file fs '("OPEN.TXT"))
    (call-with-port (fatfs-open-file fs '("OPEN.TXT") (fxior fatfs-open/read+write fatfs-open/create fatfs-open/exclusive))
      values)
    (test-error (call-with-port (fatfs-open-file fs '("OPEN.TXT") (fxior fatfs-open/read+write fatfs-open/create fatfs-open/exclusive))
                  values))
    (test-end)

    (close-fatfs fs)))

(exit (if (zero? (test-runner-fail-count (test-runner-get))) 0 1))
